@@ -26,7 +26,7 @@ from typing import Any, Callable, Dict, List, Optional, Tuple
 import torch
 from torch import nn, optim
 from torch.utils.data import DataLoader
-from tqdm.auto import tqdm
+from tqdm import tqdm
 import numpy as np
 
 @dataclass
@@ -40,6 +40,25 @@ class TrainerConfig:
         - Gradient accumulation: Train with larger effective batch sizes
         - Gradient clipping: Prevent exploding gradients
         - Non-blocking transfers: Automatic async data movement
+    
+    TensorBoard Features (Minimal CPU overhead):
+        - Real-time loss and metrics visualization (logs once per epoch)
+        - Learning rate tracking (logs once per epoch)
+        - Gradient histogram logging (optional, disabled by default - adds ~5-10% overhead)
+        - Model graph visualization (optional, logs once - negligible overhead)
+        - PyTorch Profiler integration (optional, disabled by default - adds ~10-15% overhead)
+          * GPU/CPU utilization tracking
+          * Memory usage profiling
+          * Operation timing analysis
+          * Bottleneck identification
+          * Auto-installs torch-tb-profiler plugin when enabled
+        
+        Performance Impact (typical usage):
+        - TensorBoard enabled (default): < 1% overhead (only epoch-end logging)
+        - With gradient logging: ~5-10% overhead (histogram computation is expensive)
+        - With profiler: ~10-15% overhead (detailed tracing is expensive)
+        
+        Recommendation: Keep defaults (tensorboard=True, gradients=False, profiler=False)
     """
     device: Optional[torch.device] = None
     epochs: int = 10
@@ -58,6 +77,18 @@ class TrainerConfig:
     initial_lr: Optional[float] = None
     print_every: int = 1
     seed: Optional[int] = 42
+    # TensorBoard settings
+    use_tensorboard: bool = True  # Enable TensorBoard logging
+    tensorboard_log_dir: str = "runs"  # TensorBoard log directory
+    tensorboard_log_graph: bool = True  # Log model graph
+    tensorboard_log_gradients: bool = False  # Log gradient histograms (can be slow)
+    tensorboard_comment: str = ""  # Experiment comment/name
+    # PyTorch Profiler settings (integrates with TensorBoard)
+    use_profiler: bool = False  # Enable PyTorch Profiler (performance analysis)
+    profiler_schedule_wait: int = 1  # Number of steps to skip at start
+    profiler_schedule_warmup: int = 1  # Number of steps for warmup
+    profiler_schedule_active: int = 3  # Number of steps to profile
+    profiler_schedule_repeat: int = 2  # Number of times to repeat profiling cycle
 
 
 class Callback:
@@ -131,6 +162,208 @@ class CheckpointCallback(Callback):
             trainer._save_checkpoint(filepath)
 
 
+class TensorBoardCallback(Callback):
+    """
+    TensorBoard logging callback for real-time training visualization.
+    
+    Features:
+        - Real-time loss and metrics visualization
+        - Learning rate tracking
+        - Gradient histogram logging (optional)
+        - Model graph visualization (optional)
+        - PyTorch Profiler integration (optional)
+    
+    Usage:
+        Open TensorBoard in your browser: tensorboard --logdir=runs
+        View profiler traces in the "PYTORCH_PROFILER" tab
+    """
+    
+    @staticmethod
+    def _is_torch_tb_profiler_available():
+        """Check if torch-tb-profiler is installed."""
+        try:
+            import torch_tb_profiler
+            return True
+        except ImportError:
+            return False
+    
+    @staticmethod
+    def _install_torch_tb_profiler():
+        """Auto-install torch-tb-profiler package."""
+        import subprocess
+        import sys
+        
+        try:
+            subprocess.check_call(
+                [sys.executable, "-m", "pip", "install", "-q", "torch-tb-profiler"],
+                stdout=subprocess.DEVNULL,
+                stderr=subprocess.PIPE
+            )
+            return True
+        except subprocess.CalledProcessError:
+            return False
+    
+    def __init__(
+        self,
+        log_dir: str = "runs",
+        comment: str = "",
+        log_graph: bool = True,
+        log_gradients: bool = False,
+        use_profiler: bool = False,
+        profiler_schedule_wait: int = 1,
+        profiler_schedule_warmup: int = 1,
+        profiler_schedule_active: int = 3,
+        profiler_schedule_repeat: int = 2,
+    ):
+        try:
+            from torch.utils.tensorboard import SummaryWriter
+            import datetime
+            
+            # Auto-install torch-tb-profiler if profiler is enabled
+            if use_profiler and not self._is_torch_tb_profiler_available():
+                print("\n📦 torch-tb-profiler not found, installing...")
+                if self._install_torch_tb_profiler():
+                    print("✅ torch-tb-profiler installed - PYTORCH_PROFILER tab will be available\n")
+                else:
+                    print("⚠️  Auto-install failed. Manual install: pip install torch-tb-profiler\n")
+            
+            # Create unique run name with timestamp
+            timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
+            run_name = f"{timestamp}"
+            if comment:
+                run_name = f"{comment}_{run_name}"
+            
+            self.log_dir = os.path.join(log_dir, run_name)
+            self.writer = SummaryWriter(log_dir=self.log_dir, comment=comment)
+            self.log_graph = log_graph
+            self.log_gradients = log_gradients
+            self.graph_logged = False
+            self.use_profiler = use_profiler
+            self.profiler = None
+            self.profiler_schedule_wait = profiler_schedule_wait
+            self.profiler_schedule_warmup = profiler_schedule_warmup
+            self.profiler_schedule_active = profiler_schedule_active
+            self.profiler_schedule_repeat = profiler_schedule_repeat
+            
+            # Build logging info
+            features = ["Loss/Metrics/LR"]
+            if log_graph:
+                features.append("Model Graph")
+            if log_gradients:
+                features.append("Gradients")
+            if use_profiler:
+                features.append(f"Profiler(wait={profiler_schedule_wait},warmup={profiler_schedule_warmup},active={profiler_schedule_active},repeat={profiler_schedule_repeat})")
+            
+            print(f"\n📊 TensorBoard: {', '.join(features)}")
+            print(f"   Dir: {self.log_dir}")
+            print(f"   View: tensorboard --logdir={log_dir}")
+            if use_profiler and self._is_torch_tb_profiler_available():
+                print(f"   Profiler: Look for PYTORCH_PROFILER tab\n")
+            else:
+                print()
+            
+        except ImportError:
+            print("\n⚠️  TensorBoard not available. Install with: pip install tensorboard")
+            self.writer = None
+    
+    def on_epoch_begin(self, trainer: "Trainer", epoch: int):
+        """Initialize PyTorch Profiler at start of first epoch"""
+        if self.writer is None:
+            return
+        
+        # Initialize profiler on first epoch
+        if self.use_profiler and self.profiler is None and epoch == 1:
+            try:
+                # Profiler with TensorBoard integration
+                schedule = torch.profiler.schedule(
+                    wait=self.profiler_schedule_wait,
+                    warmup=self.profiler_schedule_warmup,
+                    active=self.profiler_schedule_active,
+                    repeat=self.profiler_schedule_repeat,
+                )
+                
+                self.profiler = torch.profiler.profile(
+                    schedule=schedule,
+                    on_trace_ready=torch.profiler.tensorboard_trace_handler(self.log_dir),
+                    record_shapes=True,
+                    profile_memory=True,
+                    with_stack=True,
+                )
+                
+                self.profiler.__enter__()
+                print(f"✅ Profiler started → {self.log_dir}\n")
+                
+            except Exception as e:
+                print(f"⚠️  Profiler init failed: {e}\n")
+                self.profiler = None
+                self.use_profiler = False
+    
+    def on_batch_end(self, trainer: "Trainer", batch: int, logs: Dict[str, Any]):
+        """Step profiler (only when enabled) - optimized for minimal CPU overhead"""
+        # Early exit if writer is not available
+        if self.writer is None:
+            return
+        
+        # Only step profiler if it's active (no unnecessary work otherwise)
+        # This check is fast - just a None comparison
+        if self.profiler is not None:
+            self.profiler.step()  # Required for profiler to work correctly
+    
+    def on_epoch_end(self, trainer: "Trainer", epoch: int, logs: Dict[str, Any]):
+        """Log metrics, losses, and learning rate at epoch end"""
+        if self.writer is None:
+            return
+        
+        # Log all metrics and losses
+        for key, value in logs.items():
+            if isinstance(value, (int, float)):
+                # Separate train and val metrics
+                if key.startswith("train_"):
+                    metric_name = key.replace("train_", "")
+                    self.writer.add_scalar(f"Train/{metric_name}", value, epoch)
+                elif key.startswith("val_"):
+                    metric_name = key.replace("val_", "")
+                    self.writer.add_scalar(f"Validation/{metric_name}", value, epoch)
+                else:
+                    self.writer.add_scalar(f"Metrics/{key}", value, epoch)
+        
+        # Log learning rate
+        try:
+            current_lr = trainer.optimizer.param_groups[0]['lr']
+            self.writer.add_scalar("Learning_Rate/lr", current_lr, epoch)
+        except (AttributeError, IndexError, KeyError):
+            pass
+        
+        # Log gradient histograms (if enabled)
+        if self.log_gradients:
+            try:
+                for name, param in trainer.model.named_parameters():
+                    if param.grad is not None:
+                        self.writer.add_histogram(f"Gradients/{name}", param.grad, epoch)
+                        self.writer.add_histogram(f"Weights/{name}", param, epoch)
+            except Exception:
+                pass  # Gradient logging is optional
+        
+        # Flush to disk
+        self.writer.flush()
+    
+    def __del__(self):
+        """Close writer and profiler on cleanup"""
+        # Close profiler first
+        if hasattr(self, 'profiler') and self.profiler is not None:
+            try:
+                self.profiler.__exit__(None, None, None)
+            except Exception:
+                pass
+        
+        # Close writer
+        if hasattr(self, 'writer') and self.writer is not None:
+            try:
+                self.writer.close()
+            except Exception:
+                pass
+
+
 class Trainer:
     def __init__(
         self,
@@ -145,6 +378,22 @@ class Trainer:
         self.metrics = metrics or {}
         self.callbacks = callbacks or []
         self.helper = Helper()
+        self.gpu_augmentation = None  # Detected lazily in fit()
+        
+        # Auto-initialize TensorBoard callback if enabled
+        if config.use_tensorboard:
+            tb_callback = TensorBoardCallback(
+                log_dir=config.tensorboard_log_dir,
+                comment=config.tensorboard_comment,
+                log_graph=config.tensorboard_log_graph,
+                log_gradients=config.tensorboard_log_gradients,
+                use_profiler=config.use_profiler,
+                profiler_schedule_wait=config.profiler_schedule_wait,
+                profiler_schedule_warmup=config.profiler_schedule_warmup,
+                profiler_schedule_active=config.profiler_schedule_active,
+                profiler_schedule_repeat=config.profiler_schedule_repeat,
+            )
+            self.callbacks.append(tb_callback)
 
         self.device = config.device or (torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu"))
         
@@ -164,25 +413,17 @@ class Trainer:
         self.scaler = torch.cuda.amp.GradScaler(enabled=amp_enabled)
         
         # Print AMP status
+        device_name = torch.cuda.get_device_name(0) if torch.cuda.device_count() > 0 else "CUDA"
         if amp_enabled:
-            print("\n" + "="*60)
-            print("⚡ MIXED PRECISION TRAINING ENABLED ⚡")
-            print("="*60)
-            print("Using: Automatic Mixed Precision (AMP)")
-            print("Precision: FP16 for forward/backward, FP32 for weights")
-            print(f"Device: {torch.cuda.get_device_name(0) if torch.cuda.device_count() > 0 else 'CUDA'}")
-            print("Benefits:")
-            print("  • 2-3x faster training")
-            print("  • 50% less GPU memory")
-            print("  • No accuracy loss (with gradient scaling)")
-            print("="*60 + "\n")
+            print(f"\n⚡ AMP (FP16): {device_name} | 2-3x faster, 50% less memory\n")
         elif self.device.type == "cuda" and not config.amp:
-            print("\nℹ️  Mixed precision (AMP) is disabled.")
-            print("   Enable for 2-3x speedup: TrainerConfig(amp=True)\n")
+            print(f"\nℹ️  AMP disabled | Enable for 2-3x speedup: TrainerConfig(amp=True)\n")
         else:
             print(f"\nℹ️  Training on {self.device.type.upper()}")
             if self.device.type == "cpu":
-                print("   Note: Mixed precision (AMP) not available on CPU\n")
+                print("   (AMP not available on CPU)\n")
+            else:
+                print()
         
         os.makedirs(self.config.save_dir, exist_ok=True)
         self._best_metric = None
@@ -215,25 +456,13 @@ class Trainer:
             return model
         
         if not compile_available:
-            print("\n⚠️  torch.compile() not available (requires PyTorch 2.0+)")
-            print("   Current PyTorch version:", torch.__version__)
-            print("   Upgrade: pip install torch>=2.0.0\n")
+            print(f"\n⚠️  torch.compile() unavailable (PyTorch {torch.__version__} < 2.0)\n")
             return model
         
         # Compile the model
         try:
-            print("\n" + "="*60)
-            print("🔥 MODEL COMPILATION ENABLED 🔥")
-            print("="*60)
-            print("Using: torch.compile() [PyTorch 2.0+]")
-            print(f"Mode: {self.config.compile_mode}")
-            print(f"Device: {torch.cuda.get_device_name(0) if torch.cuda.device_count() > 0 else 'CUDA'}")
-            print("Benefits:")
-            print("  • 20-50% faster training")
-            print("  • Graph optimizations (operator fusion, etc.)")
-            print("  • Automatic kernel selection")
-            print("Note: First epoch will be slower (compilation overhead)")
-            print("="*60 + "\n")
+            device_name = torch.cuda.get_device_name(0) if torch.cuda.device_count() > 0 else "CUDA"
+            print(f"\n🔥 torch.compile(mode={self.config.compile_mode}): {device_name} | 20-50% faster (1st epoch slower)\n")
             
             compiled_model = torch.compile(
                 model,
@@ -243,8 +472,7 @@ class Trainer:
             return compiled_model
             
         except Exception as e:
-            print(f"\n⚠️  Model compilation failed: {e}")
-            print("   Continuing with non-compiled model\n")
+            print(f"\n⚠️  Compilation failed: {str(e)[:100]}... | Using non-compiled model\n")
             return model
 
     def _save_checkpoint(self, path: str):
@@ -281,25 +509,52 @@ class Trainer:
 
         pbar = tqdm(enumerate(dataloader), total=len(dataloader), desc=("Train" if is_train else "Eval"))
         self.optimizer.zero_grad()
+        
+        # Track if we've hit a torch.compile error
+        compile_failed = False
 
         for batch_idx, (X, y) in pbar:
-            for cb in self.callbacks:
-                cb.on_batch_begin(self, batch_idx)
+            # Skip batch_begin callbacks if list is empty (performance optimization)
+            if self.callbacks:
+                for cb in self.callbacks:
+                    cb.on_batch_begin(self, batch_idx)
 
             # Use non_blocking=True to enable asynchronous data transfer
             # This allows CPU to continue preparing next batch while GPU transfers current batch
             X = X.to(self.device, non_blocking=True)
             y = y.to(self.device, non_blocking=True)
+            
+            # Apply GPU augmentation if enabled (training only)
+            if is_train and self.gpu_augmentation is not None:
+                X = self.gpu_augmentation(X)
+            
             batch_size = X.shape[0]
 
-            with torch.set_grad_enabled(is_train):
-                if self.config.amp and self.device.type == "cuda":
-                    with torch.cuda.amp.autocast():
+            try:
+                with torch.set_grad_enabled(is_train):
+                    if self.config.amp and self.device.type == "cuda":
+                        # Use modern torch.amp.autocast API (PyTorch 1.10+)
+                        # Falls back to torch.cuda.amp.autocast for older versions
+                        try:
+                            autocast_context = torch.amp.autocast('cuda')
+                        except AttributeError:
+                            autocast_context = torch.cuda.amp.autocast()
+                        
+                        with autocast_context:
+                            logits = self.model(X)
+                            loss = self.loss_fn(logits, y)
+                    else:
                         logits = self.model(X)
                         loss = self.loss_fn(logits, y)
-                else:
-                    logits = self.model(X)
-                    loss = self.loss_fn(logits, y)
+            except Exception as e:
+                # Catch torch.compile errors and provide helpful feedback
+                error_msg = str(e)
+                if 'Dynamo' in error_msg or 'FakeTensor' in error_msg or 'compile' in error_msg.lower():
+                    if not compile_failed:
+                        compile_failed = True
+                        print(f"\n⚠️  torch.compile() error: {error_msg[:150]}...")
+                        print("   Fix: Check model architecture OR set compile_model=False\n")
+                raise
 
             # normalize loss across accumulation steps
             loss_value = loss.detach().item() if isinstance(loss, torch.Tensor) else float(loss)
@@ -344,8 +599,10 @@ class Trainer:
                 **{name: metric_sums[name] / total_samples for name in metric_sums},
             }
 
-            for cb in self.callbacks:
-                cb.on_batch_end(self, batch_idx, logs)
+            # Batch-end callbacks (only if callbacks exist)
+            if self.callbacks:
+                for cb in self.callbacks:
+                    cb.on_batch_end(self, batch_idx, logs)
 
             pbar.set_postfix({k: f"{v:.4f}" for k, v in logs.items()})
 
@@ -370,10 +627,26 @@ class Trainer:
         epochs = epochs or self.config.epochs
         stop_training = False
         early_stopper = next((c for c in self.callbacks if isinstance(c, EarlyStopping)), None)
+        
+        # Auto-detect GPU augmentation from train_loader's attached ImageDataLoader
+        # EAFP: Easier to Ask for Forgiveness xx Permission (Pythonic!)
+        if self.gpu_augmentation is None:
+            try:
+                # Try to get ImageDataLoader instance from train_loader
+                image_data_loader = getattr(train_loader, '_image_data_loader', None)
+                if image_data_loader is not None and image_data_loader.use_gpu_augmentation:
+                    self.gpu_augmentation = image_data_loader.get_gpu_augmentation()
+                    if self.gpu_augmentation is not None:
+                        print("\n🎨 GPU Augmentation: Auto-detected from train_loader (training only)\n")
+            except (AttributeError, TypeError):
+                # Not from ImageDataLoader or GPU augmentation not available
+                pass
 
         for epoch in range(1, epochs + 1):
-            for cb in self.callbacks:
-                cb.on_epoch_begin(self, epoch)
+            # Epoch-begin callbacks
+            if self.callbacks:
+                for cb in self.callbacks:
+                    cb.on_epoch_begin(self, epoch)
 
             train_logs = self._run_one_epoch(train_loader, epoch, training=True)
             val_logs: Dict[str, float] = {}
@@ -390,9 +663,10 @@ class Trainer:
                 or None
             )
 
-            # notify callbacks
-            for cb in self.callbacks:
-                cb.on_epoch_end(self, epoch, logs)
+            # Epoch-end callbacks (metrics logging, checkpointing, etc.)
+            if self.callbacks:
+                for cb in self.callbacks:
+                    cb.on_epoch_end(self, epoch, logs)
 
             # update history
             epoch_record = {"epoch": epoch, **logs}
@@ -409,7 +683,7 @@ class Trainer:
 
         if val_loader is not None:
             try:
-                self.helper.plot_train_test_loss(self._history)
+                self.helper.plotTrainTestLoss(self._history)
             except Exception as e:
                 print(f"Could not plot losses: {e}")
         return self._history
