@@ -1,9 +1,16 @@
 # LayerZero/__init__.py
-__version__ = "0.1.11"
+__version__ = "0.4.6"
 
-from .ImageDataLoader import ImageDataLoader
+from .ImageDataLoader import ImageDataLoader, ImageLoaderConfig
 from .Helper import Helper
-from .Trainer import Trainer, TrainerConfig
+from .Trainer import (
+    Trainer, 
+    TrainerConfig, 
+    Callback, 
+    EarlyStopping, 
+    CheckpointCallback, 
+    TensorBoardCallback
+)
 from .AugmentationMode import AugmentationMode
 
 # Kornia helper utilities (always available)
@@ -19,10 +26,15 @@ from .KorniaHelper import (
 try:
     from .GPUAugmentation import GPUAugmentation, HybridAugmentation
     __all__ = [
-        "ImageDataLoader", 
+        "ImageDataLoader",
+        "ImageLoaderConfig",
         "Helper", 
         "Trainer", 
-        "TrainerConfig", 
+        "TrainerConfig",
+        "Callback",
+        "EarlyStopping",
+        "CheckpointCallback",
+        "TensorBoardCallback",
         "AugmentationMode",
         "GPUAugmentation", 
         "HybridAugmentation",
@@ -34,10 +46,15 @@ try:
     ]
 except ImportError:
     __all__ = [
-        "ImageDataLoader", 
+        "ImageDataLoader",
+        "ImageLoaderConfig",
         "Helper", 
         "Trainer", 
-        "TrainerConfig", 
+        "TrainerConfig",
+        "Callback",
+        "EarlyStopping",
+        "CheckpointCallback",
+        "TensorBoardCallback",
         "AugmentationMode",
         "KorniaHelper",
         "is_kornia_available",
